/*******************************************************************************
 *  emeter-gpio-iic-eeprom.c -
 *
 * Copyright (C) 2015 Texas Instruments Incorporated - http://www.ti.com/ 
 * ALL RIGHTS RESERVED 
 *
 ******************************************************************************/

#include <inttypes.h>
#include <stdbool.h>
#if defined(__MSP430__)
#include <msp430.h>
#endif

#include "emeter-template.h"
#include "emeter-iic-eeprom.h"

/* These routines support a single I2C EEPROM or FRAM attached to
   two pins of a GPIO port on an MSP430.

    Please note:

    These routines have been tested with an Atmel 2401. They
    are intended to work with a range of I2C interface serial
    EEPROMs, but should be thoroughly tested against any
    other EEPROM before serious use.

    Since the EEPROM will be operating at a low voltage, when
    connected to the MSP430, this code is designed to operate
    at the worst case timing for the Atmel chips, at their
    lowest recommended supply voltage, and with the MSP430
    running at 8MHz. They could be tuned for faster operation
    in other environments. */

#if defined(EXTERNAL_EEPROM_SUPPORT)  &&  defined(EEPROM_SDA_BIT)  &&  defined(EEPROM_SCL_BIT)

#define MAX_IIC_TRIES 200

#define EEPROM_SLAVE_ADDRESS    0x50

#if defined(EEPROM_IS_REALLY_FRAM)
/* No delay is needed for FRAMs */
#define delay(x) /**/
#else
/* This delay is needed for EEPROMs */
static void delay(int i)
{
    for (  ;  i;  i--)
    {
        __no_operation();
        __no_operation();
        __no_operation();
        __no_operation();
    }
}
#endif

static void iic_start(void)
{
    /* Send a start to the EEPROM */
    /* A start is the data bit falling while the clock is high */
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    delay(2);   /* At least ??? */
    EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
    delay(5);   /* At least 4.7us */
    EEPROM_SDA_PORT_DIR |= EEPROM_SDA_BIT;
    delay(5);   /* At least 4.7us */
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
}

static void iic_stop(void)
{
    /* Send a stop to the EEPROM */
    /* A stop is the data bit rising while the clock is high */
    EEPROM_SDA_PORT_DIR |= EEPROM_SDA_BIT;
    delay(2);   /* At least ??? */
    EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
    delay(5);   /* At least 4.7us */
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    delay(5);   /* At least 4.7us */
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
}

static int iic_send(uint8_t data)
{
    int i;
    
    /* Send 8 bits of data */
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
    for (i = 8;  i > 0;  i--)
    {
        /* The data can now change without delay */
        if ((data & 0x80))
            EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
        else
            EEPROM_SDA_PORT_DIR |= EEPROM_SDA_BIT;
        data <<= 1;
        /* Pulse the clock high while the data bit is steady */
        delay(5);   /* At least 4.7us */
        EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
        delay(5);   /* At least 4.7us */
        EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
    }
    /* Check the acknowledgement from the EEPROM */
    /* Pulse the clock high, and see what the device says while the clock is high */
    delay(5);
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    delay(5);
    EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
    delay(5);
    i = EEPROM_SDA_PORT_IN & EEPROM_SDA_BIT;
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
    delay(4);
    return i;
}

static uint8_t iic_receive(int ok)
{
    uint8_t data;
    int i;

    /* Get 8 bits of data */
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;     /* Input */
    data = 0;
    for (i = 8;  i > 0;  i--)
    {
        EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
        delay(5);   /* At least 4.7us */
        data <<= 1;
        if ((EEPROM_SDA_PORT_IN & EEPROM_SDA_BIT))
            data |= 0x01;
        EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
        delay(5);   /* At least 4.7us */
    }
    /* Send the acknowledgement to the EEPROM */
    if (ok)
        EEPROM_SDA_PORT_DIR |= EEPROM_SDA_BIT;
    else
        EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    delay(4);
    EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
    delay(4);
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    return data;
}

static int test_sda(void)
{
    int i;
  
    iic_stop();
    EEPROM_SDA_PORT_DIR &= ~EEPROM_SDA_BIT;
    delay(4);
    for (i = 16;  i > 0;  i--)
    {
        delay(5);
        if (!(EEPROM_SDA_PORT_IN & EEPROM_SDA_BIT))
            break;
    }
    return i;
}

int eeprom_read(eeprom_addr_t address, uint8_t buf[], int unsigned size)
{
    int i;
    int j;
    uint8_t slave_address;

    if (size == 0)
        return 0;

    for (i = 0;  i < MAX_IIC_TRIES;  ++i)
    {
        if (i)
        {
            /* Read false, retry */
            if (test_sda())
                continue;
        }
        iic_start();
#if EEPROM_ADDRESS_BITS_IN_I2C_ADDR > 0  ||  EEPROM_CHIP_SEL_BITS_IN_I2C_ADDR > 0
        slave_address = (EEPROM_SLAVE_ADDRESS | (address >> (EEPROM_ADDRESS_BITS - EEPROM_ADDRESS_BITS_IN_I2C_ADDR))) << 1;
#else
        slave_address = EEPROM_SLAVE_ADDRESS << 1;
#endif
        if (iic_send(slave_address))
            continue;
#if EEPROM_ADDRESS_BYTES == 2
        if (iic_send(addr >> 8))
            continue;
#endif
        if (iic_send(address & 0xFF))
            continue;

        iic_start();
        if (iic_send(slave_address | 1))
            continue;
        for (j = 0;  j < size;  j++)
            buf[j] = iic_receive(true);
        buf[j] = iic_receive(false);
        iic_stop();
        return size;
    }
    iic_stop();
    return EEPROM_RESULT_NACK;
}

int eeprom_write(eeprom_addr_t address, const uint8_t buf[], unsigned int size)
{
    int i;
    int j;
    unsigned int ptr;
    unsigned int len_data;
    uint8_t *q;
    uint8_t slave_address;

    if (size == 0)
        return 0;

    /* Write an arbitrarily sized buffer, in EEPROM page limited chunks */
    q = (uint8_t *) buf;
    for (ptr = 0;  ptr < size;  ptr += len_data)
    {
        /* We need to limit the size of this write to remain within the current EEPROM page */
        len_data = EEPROM_PAGE_SIZE - (address & (EEPROM_PAGE_SIZE - 1));
        if (len_data > (size - ptr))
            len_data = size - ptr;
        for (i = 0;  i < MAX_IIC_TRIES;  ++i)
        {
            if (i)
            {
                /* Write false, retry */
                if (test_sda())
                    continue;
            }

            iic_start();
#if EEPROM_ADDRESS_BITS_IN_I2C_ADDR > 0  ||  EEPROM_CHIP_SEL_BITS_IN_I2C_ADDR > 0
            slave_address = (EEPROM_SLAVE_ADDRESS | (address >> (EEPROM_ADDRESS_BITS - EEPROM_ADDRESS_BITS_IN_I2C_ADDR))) << 1;
#else
            slave_address = EEPROM_SLAVE_ADDRESS << 1;
#endif
            if (iic_send(slave_address))
                continue;
#if EEPROM_ADDRESS_BYTES == 2
            if (iic_send(address >> 8))
                continue;
#endif
            if (iic_send(address & 0xFF))
                continue;
            for (j = 0;  j < len_data;  j++)
            {
                if (iic_send(q[j]))
                    break;
            }
            break;
            //iic_stop();
        }
        iic_stop();
        if (i >= MAX_IIC_TRIES)
            return -1;
        address += len_data;
        q += len_data;
    }
    return size;
}

int external_eeprom_init(void)
{
    int i;

    /* While idling, the EEPROM clock should be low */
    EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
    /* If we happen to have restarted in the middle of a read from
       the EEPROM/FRAM, we need to regain control of the device. If we
       give it enough clocks, and do no acknowledge it we should get out
       of any odd conditions. Then we do a stop, and we should be OK. If
       the device was mid write when the restart occurred we cannot really
       act in a clean way. */
    delay(5);   /* At least 4.7us */
    for (i = 0;  i < 16;  i++)
    {
        EEPROM_SCL_PORT_DIR &= ~EEPROM_SCL_BIT;
        delay(5);   /* At least 4.7us */
        EEPROM_SCL_PORT_DIR |= EEPROM_SCL_BIT;
        delay(5);   /* At least 4.7us */
    }
    iic_stop();
    return 0;
}
#endif
